/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart;

import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getNodeBounds;

import java.util.List;

import org.conqat.ide.commons.ui.swt.FontUtils;
import org.eclipse.draw2d.ChopboxAnchor;
import org.eclipse.draw2d.ConnectionAnchor;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPolicy;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.fortiss.tooling.base.layout.DefaultLayoutConstants;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.editpart.figure.TransparentLabel;
import org.fortiss.tooling.base.ui.editpart.policy.DirectConnectionAndReconnectionEditPolicy;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * The base class for graphical edit parts which encapsulate nodes (i.e. the
 * main modeling elements). All of them have a name label and an activation
 * mode. This EditPart is intended to be used in allocation diagrams which
 * allow the connection of elements from two different GEF based diagram editors.
 * 
 * @param <T>
 *            the type of the model element wrapped.
 * 
 * @author eder
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public abstract class AllocationElementEditPartBase<T extends ILayoutedModelElement & INamedCommentedElement>
		extends PositionedEditPartBase<T> {

	/** The default font for titles. */
	public static Font DEFAULT_TITLE_FONT = FontUtils.derive(JFaceResources.getDefaultFont(), 0,
			SWT.BOLD);

	/** The allocation object. (e.g. a deployment) */
	private EObject allocationObject;

	/** Constructor. */
	protected AllocationElementEditPartBase(T modelElement, EObject allocationObject) {
		super(modelElement);
		this.allocationObject = allocationObject;
	}

	/**
	 * If {@link T} is the source of a allocation connection this method has to be implemented.
	 * Other wise return an empty list.
	 * 
	 * @return a list of all outgoing connections of this element
	 */
	@SuppressWarnings("javadoc")
	protected abstract List<?> getModelSourceConnectionsInternal();

	/**
	 * If {@link T} is the target of an allocation connection this method has to be implemented.
	 * Other wise return an empty list.
	 * 
	 * @return a list of all incoming connections of this element
	 */
	@SuppressWarnings("javadoc")
	protected abstract List<?> getModelTargetConnectionsInternal();

	/** {@inheritDoc} */
	@Override
	protected Rectangle determineLabelFigureBounds() {
		Rectangle textBounds = labelFigure.getTextBounds();
		Rectangle labelBounds = determineBaseFigureBounds();
		labelBounds.y += (labelBounds.height - textBounds.height) / 2;
		labelBounds.x += 2 * DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
		labelBounds.width -= 4 * DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
		labelBounds.height = textBounds.height * 2;
		return labelBounds;
	}

	/** Creates the label figure. */
	@Override
	public Label createLabelFigure() {
		Label nameLabel = new TransparentLabel();
		nameLabel.setForegroundColor(getForegroundColor());
		nameLabel.setFont(DEFAULT_TITLE_FONT);
		nameLabel.setTextAlignment(PositionConstants.LEFT | PositionConstants.TOP);
		nameLabel.setTextPlacement(PositionConstants.EAST);
		nameLabel.setIcon(IModelElementHandlerService.INSTANCE.getModelElementHandler(getModel())
				.getIcon(getModel()));
		return nameLabel;
	}

	/** {@inheritDoc} */
	@Override
	protected Rectangle determineBaseFigureBounds() {
		T model2 = getModel();
		Rectangle r = getNodeBounds(model2);
		return r;
	}

	/** {@inheritDoc} */
	@Override
	protected ConnectionAnchor defaultConnectionAnchor() {
		return new ChopboxAnchor(getFigure());
	}

	/** {@inheritDoc} */
	@Override
	protected void createEditPolicies() {
		installEditPolicy(EditPolicy.GRAPHICAL_NODE_ROLE,
				new DirectConnectionAndReconnectionEditPolicy());

	}

	/** {@inheritDoc} */
	@Override
	protected final List<?> getModelSourceConnections() {
		return getModelSourceConnectionsInternal();
	}

	/** {@inheritDoc} */
	@Override
	protected List<?> getModelTargetConnections() {
		return getModelTargetConnectionsInternal();
	}

	/** {@inheritDoc} */
	@Override
	protected String determineLabelFigureText() {
		return getModel().getName();
	}

	/** Returns allocationObject. */
	public EObject getAllocationObject() {
		return allocationObject;
	}

}
