/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor.EReferenceScope;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s that manages {@link EStructuralFeature}s
 * contained by a {@link IAnnotatedSpecification}.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 484315167A409237789D478428543CAF
 */
public abstract class EStructuralFeatureValueProviderBase<T extends IAnnotatedSpecification>
		extends ValueProviderBase<T> {

	/**
	 * Map: Instance key -> {@link EStructuralFeatureDescriptor}s of the
	 * {@link IAnnotatedSpecification} managed by this {@link IAnnotationValueProvider}.
	 */
	final private Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap;

	/**
	 * Flag if multiple {@link EStructuralFeature}s are managed by this
	 * {@link IAnnotationValueProvider}.
	 */
	protected boolean handlesMultipleEstructuralFeatures;

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by a single {@link EAttribute}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EAttribute eAttribute) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();
		structuralFeatureDescriptorMap.put(DEFAULT_KEY,
				new EStructuralFeatureDescriptor(eAttribute));

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by a single {@link EReference}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			EReference eReference, EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();
		structuralFeatureDescriptorMap.put(DEFAULT_KEY, new EStructuralFeatureDescriptor(
				eReference, eReferenceScope));

		handlesMultipleEstructuralFeatures = false;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by a number of {@link EStructuralFeature}s contained in a single
	 * {@link EClass}. The mapping of instance keys to the corresponding {@link EStructuralFeature}s
	 * is explicitly provided in a {@link Map}.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap) {

		super(annotatedSpecificationEClass);

		this.structuralFeatureDescriptorMap = structuralFeatureDescriptorMap;

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a {@link IAnnotatedSpecification} whose
	 * storage is implemented by all {@link EStructuralFeature}s defined locally in a single
	 * {@link EClass} (i.e., not considering inherited attributes). The instance keys, and the
	 * mapping to the corresponding {@link EStructuralFeature}s are automatically derived from the
	 * underlying {@link EClass} (instance keys = field names of {@link EStructuralFeature}).
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass) {

		this(annotatedSpecificationEClass, false,
				EStructuralFeatureDescriptor.EReferenceScope.MODEL);
	}

	/**
	 * Generalizes the interface of {@link #EStructuralFeatureValueProviderBase(EClass)} with the
	 * following two parameters:
	 * 
	 * @param inheritedFeatures
	 *            Flag if {@link EStructuralFeature}s that are not defined locally in
	 *            {@code annotatedSpecificationEClass}, but that are inherited, should be included.
	 * @param eReferenceScope
	 *            {@link EReferenceScope} to be offered to the user to select values for
	 *            {@link EReference}s.
	 */
	public EStructuralFeatureValueProviderBase(EClass annotatedSpecificationEClass,
			boolean inheritedFeatures, EStructuralFeatureDescriptor.EReferenceScope eReferenceScope) {

		super(annotatedSpecificationEClass);

		structuralFeatureDescriptorMap = new HashMap<String, EStructuralFeatureDescriptor>();

		EList<EStructuralFeature> featureList;
		if(inheritedFeatures) {
			featureList = annotatedSpecificationEClass.getEAllStructuralFeatures();
		} else {
			featureList = annotatedSpecificationEClass.getEStructuralFeatures();
		}

		for(EStructuralFeature structuralFeature : featureList) {
			// - In case there is only one feature, it is managed by the DEFAULT_KEY (see
			// documentation of DEFAULT_KEY).
			// - Otherwise, use the feature's name as instance key.
			String instanceKey =
					featureList.size() == 1 ? DEFAULT_KEY : structuralFeature.getName();
			if(structuralFeature instanceof EAttribute) {
				structuralFeatureDescriptorMap.put(instanceKey, new EStructuralFeatureDescriptor(
						(EAttribute)structuralFeature));
			} else if(structuralFeature instanceof EReference) {
				structuralFeatureDescriptorMap.put(instanceKey, new EStructuralFeatureDescriptor(
						(EReference)structuralFeature, eReferenceScope));
			}
		}

		handlesMultipleEstructuralFeatures = true;
	}

	/**
	 * Returns the {@link EStructuralFeature} for a given {@code instanceKey}, or throws
	 * an exception, if the key is unknown.
	 * 
	 */
	public EStructuralFeature getEStructuralFeature(T specification, String instanceKey)
			throws Exception {
		EStructuralFeatureDescriptor eStructuralFeatureDescriptor =
				structuralFeatureDescriptorMap.get(instanceKey);
		EStructuralFeature structuralFeature = null;
		if(eStructuralFeatureDescriptor != null) {
			structuralFeature = eStructuralFeatureDescriptor.getEStructuralFeature(specification);
		}
		if(structuralFeature == null) {
			throw new Exception("Instance key " + instanceKey +
					" is unknown for annotation provider " +
					annotatedSpecificationEClass.getInstanceClassName() + ".");
		}

		return structuralFeature;
	}

	/** Returns the {@link EStructuralFeatureDescriptor} for a given {@code instanceKey}. */
	public EStructuralFeatureDescriptor getEStructuralFeatureDescriptor(String instanceKey) {
		return structuralFeatureDescriptorMap.get(instanceKey);
	}

	/** Returns the {@link EFactory} for the given {@link EStructuralFeature}. */
	private EFactory getEFactory(EStructuralFeature structuralFeature) {
		return structuralFeature.getEType().getEPackage().getEFactoryInstance();
	}

	/**
	 * Sets a value for a {@link IAnnotatedSpecification} from a {@link String} representation of
	 * the input.
	 */
	protected <V> void setAnnotationValueFromString(String value, T specification,
			String instanceKey) throws Exception {

		EStructuralFeature structuralFeature = getEStructuralFeature(specification, instanceKey);

		if(structuralFeature instanceof EAttribute) {
			EFactory eFactory = getEFactory(structuralFeature);
			EDataType eAttributeType = ((EAttribute)structuralFeature).getEAttributeType();
			specification.eSet(structuralFeature, eFactory.createFromString(eAttributeType, value));
		} else {
			throw new Exception(
					"setAnnotationValueFromString() is only availabe for EAttributes. Use setAnnotationValue(U value, T specification) instead.");
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification, String instanceKey) {
		return true;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		if(value instanceof String) {
			// In case the input choice is presented in dynamically managed combo box (i.e., where
			// new value choices can be entered by the user), the empty String indicates that the
			// current value should be deleted, and a neighboring value from the set should be used.
			// In case the last occurrence of a value choice has been removed, it will no longer
			// be offered in the combo box.
			if(((String)value).isEmpty()) {
				String rval =
						determineCurrentAnnotationValueAndUpdateInputChoice((String)value,
								specification, instanceKey);
				if(rval != null) {
					value = (U)rval;
				}
			}

			setAnnotationValueFromString((String)value, specification, instanceKey);
		} else {
			if(value != null) {
				specification.eSet(getEStructuralFeature(specification, instanceKey), value);
			} else {
				// Treat 'null' as special value used to unset EStructural features. This
				// implementation choice prevents to distinguish between 'null' and unset for
				// EReferences).
				specification.eUnset(getEStructuralFeature(specification, instanceKey));
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		// See comment in setAnnotationValue(U value, T specification)
		if((value == null || value.isEmpty())) {
			String rval =
					determineCurrentAnnotationValueAndUpdateInputChoice(value, specification,
							instanceKey);
			if(rval != null) {
				value = rval;
			}
		}

		EStructuralFeature structuralFeature = getEStructuralFeature(specification, instanceKey);

		if(structuralFeature instanceof EAttribute) {
			setAnnotationValueFromString(value, specification, instanceKey);
		} else {
			throw new Exception(
					"setAnnotationValue(String, T) is not supported / has not been implemented for the annotation type " +
							structuralFeature.getEType().getName() +
							"(specification class: " +
							specification.getClass() + ").");
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return (U)specification.eGet(getEStructuralFeature(specification, instanceKey));
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification) throws Exception {
		return getAnnotationValue(specification, DEFAULT_KEY);
	}

	/**
	 * Creates a new instance of the {@link EStructuralFeature} (used to populate newly created
	 * annotation instances created in
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}).
	 */
	protected Object createStructuralFeatureInstance(String instanceKey) throws Exception {
		EStructuralFeature structuralFeature = getEStructuralFeature(null, instanceKey);

		EFactory factory = getEFactory(structuralFeature);
		if(structuralFeature instanceof EAttribute) {

			EDataType attributeEDataType = ((EAttribute)structuralFeature).getEAttributeType();

			String defaultValueLiteral = structuralFeature.getDefaultValueLiteral();
			if(defaultValueLiteral != null) {
				// Use default value if it has been specified
				return factory.createFromString(attributeEDataType, defaultValueLiteral);
			}

			// Return data types default value
			return attributeEDataType.getDefaultValue();
		} else if(structuralFeature instanceof EReference) {
			return null;
		} else {
			throw new Exception(
					"createStructuralFeatureInstance() is not supported / has not been implemented for the annotation type " +
							structuralFeature.getName() + ".");
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {

		// Create and set structural feature implementing the annotation
		Object structuralFeatureVal = createStructuralFeatureInstance(instanceKey);
		if(structuralFeatureVal != null) {
			EStructuralFeature eStructuralFeature =
					getEStructuralFeature(specification, instanceKey);
			specification.eSet(eStructuralFeature, structuralFeatureVal);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		if(handlesMultipleEstructuralFeatures) {
			return structuralFeatureDescriptorMap.keySet();
		}
		return super.getInstanceKeys(specification);
	}
}
